"""EventBridge event models"""

from __future__ import annotations

import datetime as dt
import uuid

#: pylint: disable=no-name-in-module
from pydantic import Field

from .base import AwsAccountNumber, AwsEvent, DataT

from typing import Any, Dict, Generic, List, Literal, Optional

#: pylint: enable=no-name-in-module


__all__ = ["EventBridgeResource", "EventBridgeEvent"]


## EventBridge


class EventBridgeResource(str):
    """A `str` subtype that implements the `resource_name` property"""

    @property
    def resource_name(self) -> str:
        """Resource name

        Returns
        -------
        str
            Resource name
        """
        return self.split("/")[1]


class EventBridgeEvent(AwsEvent, Generic[DataT]):
    """Event Bridge Event

    Parameters
    ----------
    version : Optional[str] = "0"
        By default, this is set to 0 (zero) in all events.
    id_ : uuid.UUID
        A Version 4 UUID that's generated for every event. You can use
        `id` to trace events as they move through rules to targets.
    detail_type : str
        Identifies, in combination with the source field, the fields and
        values that appear in the detail field.
    source : Literal["aws.events"]
        Identifies the service that generated the event; we only support
        `"aws.events"` at this point
    account : AwsAccountNumber
        12 digit AWS account number
    time : str
        The event timestamp, which can be specified by the service
        originating the event. If the event spans a time interval, the
        service can report the start time, so this value might be before
        the time the event is received.
    region : str
        AWS Region
    resources : List[EventBridgeResource]
        List of ARNs that identify the resources involved in the event
    detail : Dict[str, Any]
        A JSON object that contains information about the event.
    """

    version: Optional[str] = "0"
    id_: uuid.UUID = Field(alias="id")
    source: Literal["aws.events"]
    account: AwsAccountNumber
    time: dt.datetime
    region: str
    resources: List[EventBridgeResource]
    detail_type: str = Field(alias="detail-type")
    detail: DataT
    replay_name: Optional[str] = Field(None, alias="replay-name")

    @property
    def resource_name(self) -> str:
        """The resource name of an EventBridge event is defined as the
        resource name of its first resource

        Returns
        -------
        str
            Resource name
        """
        return self.resources[0].resource_name

    @property
    def event_key(self) -> str:
        """The event key of an EventBridge event is its resource name

        Returns
        -------
        str
            Event key
        """
        return self.resource_name

    @classmethod
    def matches_event_type(cls, event: Dict[str, Any]) -> bool:
        """Infers whether the given event is an EventBridge event

        Parameters
        ----------
        event : Dict[str, Any]
            Lambda event

        Returns
        -------
        bool
            `True` if `event` is an EventBridge event
        """
        return event.get("source") == "aws.events"



"""Defines the main `LambdaHandler` class"""

from lambda_handler.model import (
    ApiGatewayEvent,
    AwsEvent,
    DataT,
    DirectInvocationEvent,
    EventBridgeEvent,
    LambdaResponse,
    SnsEvent,
    SqsEvent,
)
from lambda_handler.types import (
    AwsEventRawCallable,
    FastApiInterface,
    LambdaContext,
    LambdaHandlerInterface,
    MangumInterface,
)
from lambda_handler.utils import OnetimeDictionary, create_wrapper, parse_lambda_event

from typing import (
    Any,
    Callable,
    Dict,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)
from typing_extensions import Concatenate, ParamSpec, TypeAlias

__all__ = ["InvalidRouteError", "ExistingRouteError", "LambdaHandler"]


Bool: TypeAlias = Union[Literal[True], Literal[False]]

_T = TypeVar("_T")
_P = ParamSpec("_P")


VALID_EVENTS = [
    "DirectInvocationEvent",
    "EventBridgeEvent",
    "SqsEvent",
    "SnsEvent",
]


class InvalidRouteError(Exception):
    """Raised when a route is not registered with the handler"""

    pass


class ExistingRouteError(Exception):
    """Raised when an existing route is overwritten by another handling
    function
    """

    pass


class NoFastAPIInstanceError(Exception):
    """Raised when an AWS API Gateway Proxy event is passed to a handler
    without a FastAPI instance
    """

    pass


class LambdaHandler(LambdaHandlerInterface):
    """AWS AWS Lambda Handler

    This class is the top-level handler which provides wrappers for module functions
    which handle AWS Lambda events of different types.

    Parameters
    ----------
    fastapi_app : Optional[FastApiInterface] = None
        Optional FastAPI instance for handling HTTP requests

    Examples
    --------
    ```python
    from lambda_handler import LambdaHandler, SnsEvent, LambdaResponse

    handler = LambdaHandler()

    @handler.sns(topic_name="MyTopic")
    def handle_mytopic(event: SnsEvent) -> LambdaResponse:
        body = frobincate()
        return LambdaResponse(status_code=200, body=body)
    ```
    """

    def __init__(self, fastapi_app: Optional[FastApiInterface] = None):

        self._events_dict: Dict[
            Tuple[str, str], AwsEventRawCallable
        ] = OnetimeDictionary()
        self._fastapi_app: Optional[FastApiInterface] = None
        self._mangum: Optional[MangumInterface] = None

        if fastapi_app is not None:
            try:
                #: pylint: disable=import-outside-toplevel
                from fastapi import FastAPI
                from mangum import Mangum

                #: pylint: enable=import-outside-toplevel
            except ImportError:
                raise ImportError(
                    "Cannot import FastAPI and Mangum! Are these installed?"
                ) from None

            cast(FastAPI, fastapi_app)
            self._fastapi_app = fastapi_app
            self._mangum = Mangum(fastapi_app)

    def __call__(self, event: Dict[str, Any], context: LambdaContext) -> Dict[str, Any]:
        return self.process_event(event, context)

    @property
    def fastapi_app(self) -> Optional[FastApiInterface]:
        """FastAPI app instance, if there is one

        Returns
        -------
        Optional[FastAPI]
            FastAPI app instance, if there is one
        """
        return self._fastapi_app

    @fastapi_app.setter
    def fastapi_app(self, fastapi_app: FastApiInterface) -> None:
        try:
            #: pylint: disable=import-outside-toplevel
            from fastapi import FastAPI
            from mangum import Mangum

            #: pylint: enable=import-outside-toplevel
        except ImportError:
            raise ImportError(
                "Cannot import FastAPI and Mangum! Are these installed?"
            ) from None
        if not isinstance(fastapi_app, FastAPI):
            raise ValueError("`fastapi_app` must be a `FastAPI` instance!")
        self._fastapi_app = fastapi_app
        self._mangum = Mangum(fastapi_app)

    @property
    def mangum(self) -> Optional[MangumInterface]:
        """Mangum handler instance, if there is one

        Returns
        -------
        Optional[MangumInterface]
            Mangum app instance, if there is one
        """
        return self._mangum

    def process_event(
        self, event: Dict[str, Any], context: LambdaContext
    ) -> Dict[str, Any]:
        """Process an event into an `AwsEvent` instance and pass it to its
        handling function

        Parameters
        ----------
        event : Dict[str, Any]
            Event dictionary
        context : LambdaContext
            Lambda context object

        Returns
        -------
        Dict[str, Any]
            Response dictionary

        Raises
        ------
        InvalidRouteError
            Raised when there is no valid route for the event
        """
        parsed_event = parse_lambda_event(event)

        if not isinstance(parsed_event, ApiGatewayEvent):

            event_type_name = type(parsed_event).__qualname__.split("[")[0]

            func = self._events_dict.get((event_type_name, parsed_event.event_key))

            if func is None:
                raise InvalidRouteError(
                    f"No valid route for event type `{event_type_name}` "
                    f"and event key `{parsed_event.event_key}`"
                )

            # The function itself parses the event so we don't need to
            # pass the parsed event

            return func(event)

        if self.mangum is None:
            raise NoFastAPIInstanceError("No FastAPI instance set on handler!")

        #: pylint: disable=not-callable
        return self.mangum(parsed_event.__root__, context)

    def add_event_func(
        self,
        event_type: Type[AwsEvent],
        event_key: str,
        func: AwsEventRawCallable,
    ) -> None:
        """Register an event function on the instance

        Parameters
        ----------
        event_type: AwsEventType
            The type of event
        event_key: str
            The event key (topic name, queue name, etc)
        func: Union[AwsEventTypedCallable[AwsEvent], AwsEventRawCallable]
            The callable to register the event for
        """
        if event_type.__qualname__ not in VALID_EVENTS:
            raise ValueError(
                f"I don't know what to do with an event of type `{event_key}`!"
            )
        event_type_name = event_type.__qualname__.split("[")[0]
        try:
            self._events_dict[(event_type_name, event_key)] = func
        except ValueError:
            existing_func = self._events_dict[(event_type_name, event_key)]
            msg = (
                "Route {} for event type {} is already registered by function {}"
            ).format(event_key, event_type_name, existing_func.__qualname__)
            raise ExistingRouteError(msg) from None

    @overload
    def direct_invocation(
        self, trigger_name: str, raw: Literal[False] = False
    ) -> Callable[
        [Callable[Concatenate[AwsEvent[DirectInvocationEvent], _P], LambdaResponse]],
        AwsEventRawCallable,
    ]:
        ...

    @overload
    def direct_invocation(
        self, trigger_name: str, raw: Literal[True]
    ) -> Callable[
        [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
        AwsEventRawCallable,
    ]:
        ...

    def direct_invocation(
        self, trigger_name: str, raw: Bool = False
    ) -> Union[
        Callable[
            [
                Callable[
                    Concatenate[AwsEvent[DirectInvocationEvent], _P], LambdaResponse
                ]
            ],
            AwsEventRawCallable,
        ],
        Callable[
            [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
            AwsEventRawCallable,
        ],
    ]:
        """Handles direct invocation events

        Parameters
        ----------
        trigger_name : str
            Name of the triggering event
        raw : bool = `False`
            Indicates that the wrapped function accepts and returns a Dict[str, Any]
            instead of a `DirectInvocationEvent` and `LambdaResponse`

        Returns
        -------
        AwsEventTypedCallable[DirectInvocationEvent]
            A callable that handles a DirectInvocationEvent, or a Dict[str, Any]

        Examples
        --------
        ```python
        from lambda_handler import (
            LambdaHandler,
            DirectInvocationEvent,
            LambdaResponse,
        )

        handler = LambdaHandler()

        @handler.direct_invocation(trigger_name="MyTrigger")
        def handle_direct_invocation(event: DirectInvocationEvent) -> LambdaResponse:
            body = frobincate()
            return LambdaResponse(status_code=200, body=body)
        ```
        """
        if raw:
            return create_wrapper(
                handler=self,
                event_type=DirectInvocationEvent,
                event_key=trigger_name,
                raw=True,
            )
        return create_wrapper(
            handler=self,
            event_type=DirectInvocationEvent,
            event_key=trigger_name,
            raw=False,
        )

    @overload
    def event_bridge(
        self, resource_name: str, raw: Literal[False] = False
    ) -> Callable[
        [Callable[Concatenate[AwsEvent[EventBridgeEvent], _P], LambdaResponse]],
        AwsEventRawCallable,
    ]:
        ...

    @overload
    def event_bridge(
        self, resource_name: str, raw: Literal[True]
    ) -> Callable[
        [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
        AwsEventRawCallable,
    ]:
        ...

    def event_bridge(
        self, resource_name: str, raw: Bool = False
    ) -> Union[
        Callable[
            [Callable[Concatenate[AwsEvent[EventBridgeEvent], _P], LambdaResponse]],
            AwsEventRawCallable,
        ],
        Callable[
            [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
            AwsEventRawCallable,
        ],
    ]:
        """Handles EventBridge events

        Parameters
        ----------
        resource_name : str
            Name of the triggering EventBridge resource
        raw : bool = `False`
            Indicates that the wrapped function accepts and returns a Dict[str, Any]
            instead of an `EventBridgeEvent` and `LambdaResponse`

        Returns
        -------
        AwsEventTypedCallable[EventBridgeEvent]
            A callable that handles a EventBridgeEvent, or a Dict[str, Any]

        Examples
        --------
        ```python
        handler = LambdaHandler()

        @handler.event_bridge(resource_name="MyResource")
        def handle_event_bridge(event: EventBridgeEvent) -> LambdaResponse:
            body = frobincate()
            return LambdaResponse(status_code=200, body=body)
        ```
        """

        return create_wrapper(
            handler=self,
            event_type=EventBridgeEvent,
            event_key=resource_name,
            raw=raw,
        )

    @overload
    def sns(
        self,
        topic_name: str,
        raw: Literal[False] = False,
    ) -> Callable[
        [Callable[Concatenate[SnsEvent[DataT], _P], LambdaResponse]],
        AwsEventRawCallable,
    ]:
        ...

    @overload
    def sns(
        self,
        topic_name: str,
        raw: Literal[True],
    ) -> Callable[
        [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
        AwsEventRawCallable,
    ]:
        ...

    def sns(
        self, topic_name: str, raw: Bool = False
    ) -> Union[
        Callable[
            [Callable[Concatenate[SnsEvent[DataT], _P], LambdaResponse]],
            AwsEventRawCallable,
        ],
        Callable[
            [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
            AwsEventRawCallable,
        ],
    ]:
        """Handles SNS events

        Parameters
        ----------
        topic_name : str
            SNS topic name
        raw : bool = `False`
            Indicates that the wrapped function accepts and returns a Dict[str, Any]
            instead of an `SnsEvent` and `LambdaResponse`

        Returns
        -------
        AwsEventTypedCallable[SnsEvent]
            A callable that handles an SnsEvent, or a Dict[str, Any]
        """

        return create_wrapper(
            handler=self,
            event_type=SnsEvent,
            event_key=topic_name,
            raw=raw,
        )

    @overload
    def sqs(
        self, queue_name: str, raw: Literal[False]
    ) -> Callable[
        [Callable[Concatenate[AwsEvent[SqsEvent], _P], LambdaResponse]],
        AwsEventRawCallable,
    ]:
        ...

    @overload
    def sqs(
        self, queue_name: str, raw: Literal[True]
    ) -> Callable[
        [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
        AwsEventRawCallable,
    ]:
        ...

    def sqs(
        self, queue_name: str, raw: Bool = False
    ) -> Union[
        Callable[
            [Callable[Concatenate[AwsEvent[SqsEvent], _P], LambdaResponse]],
            AwsEventRawCallable,
        ],
        Callable[
            [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
            AwsEventRawCallable,
        ],
    ]:
        """Handles SQS events

        Parameters
        ----------
        queue_name : str
            Name of the SQS queue
        raw : bool = `False`
            Indicates that the wrapped function accepts and returns a Dict[str, Any]
            instead of an `SqsEvent` and `LambdaResponse`

        Returns
        -------
        AwsEventTypedCallable[SqsEvent]
            A callable that handles an SqsEvent, or a Dict[str, Any]
        """

        return create_wrapper(
            handler=self,
            event_type=SqsEvent,
            event_key=queue_name,
            raw=raw,
        )



"""Type stubs and protocols"""

from lambda_handler.model import AwsEvent, LambdaResponse

from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    MutableMapping,
    Optional,
    Type,
    TypeVar,
)
from typing_extensions import Protocol, TypeAlias, TypedDict

#: pylint: disable=invalid-name, too-few-public-methods
AwsEventType = TypeVar("AwsEventType", bound=AwsEvent)


AwsEventTypedCallable = Callable[[AwsEventType], LambdaResponse]
AwsEventCallable = Callable[[Dict[str, Any]], LambdaResponse]

AwsEventRawCallable = Callable[[Dict[str, Any]], Dict[str, Any]]


class LambdaHandlerInterface(Protocol):
    """Class with an `add_event_func` method"""

    def add_event_func(
        self,
        event_type: Type[AwsEvent],
        event_key: str,
        func: Callable[[Dict[str, Any]], Dict[str, Any]],
    ):
        """Add an event function to the instance"""
        pass


LambdaEvent = Dict[str, Any]

Headers: TypeAlias = List[List[bytes]]


class Response(TypedDict):
    """Response typed dictionary"""

    status: int
    headers: Headers
    body: bytes


class LambdaCognitoIdentity(Protocol):
    """Lambda cognito identity"""

    cognito_identity_id: str
    cognito_identity_pool_id: str


class LambdaMobileClient(Protocol):
    "Lambda mobile client"

    installation_id: str
    app_title: str
    app_version_name: str
    app_version_code: str
    app_package_name: str


class LambdaMobileClientContext(Protocol):
    """Lambda mobile client context"""

    client: LambdaMobileClient
    custom: Dict[str, Any]
    env: Dict[str, Any]


class LambdaContext(Protocol):
    """Lambda context"""

    function_name: str
    function_version: str
    invoked_function_arn: str
    memory_limit_in_mb: int
    aws_request_id: str
    log_group_name: str
    log_stream_name: str
    identity: Optional[LambdaCognitoIdentity]
    client_context: Optional[LambdaMobileClientContext]

    def get_remaining_time_in_millis(self) -> int:
        """Returns the number of milliseconds left before the execution times out."""
        pass


Message: TypeAlias = MutableMapping[str, Any]
Scope: TypeAlias = MutableMapping[str, Any]
Receive: TypeAlias = Callable[[], Awaitable[Message]]
Send: TypeAlias = Callable[[Message], Awaitable[None]]


class FastApiInterface(Protocol):
    """FastAPI protocol"""

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        ...  # pragma: no cover


class MangumInterface(Protocol):
    """Callable protocol"""

    def __call__(self, event: LambdaEvent, context: LambdaContext) -> dict:
        pass



"""Main package"""

#: pylint: disable=duplicate-code, unused-import

from lambda_handler.dependencies import Dependency  # noqa: F401
from lambda_handler.handler import LambdaHandler  # noqa: F401
from lambda_handler.model import (  # noqa: F401
    DirectInvocationEvent,
    EventBridgeEvent,
    LambdaResponse,
    SnsEvent,
    SqsEvent,
)

__version__ = "0.1.0"



"""Direct Invocation models"""

from __future__ import annotations

import datetime as dt

from pydantic import validator

from .base import AwsEvent, BranchGenericModel, DataT

from typing import Any, Dict, Generic, Optional

__all__ = ["DirectInvocationDetail", "DirectInvocationEvent"]


## DirectInvocation


class DirectInvocationDetail(BranchGenericModel, Generic[DataT]):
    """Direct Invocation details

    Parameters
    ----------
    trigger : str
        Invocation trigger
    body : DataT
        Body
    meta : Optional[str]
        Metadata
    """

    trigger: str
    body: DataT
    meta: Optional[str]


class DirectInvocationEvent(AwsEvent, Generic[DataT]):
    """Direct Invocation event

    Parameters
    ----------
    direct_invocation : DirectInvocationDetail
        `DirectInvocationDetail` instance
    time_stamp : dt.datetime
        Invocation timestamp
    source : str
        Invocation source
    """

    direct_invocation: DirectInvocationDetail[DataT]
    time_stamp: dt.datetime
    source: str

    @validator("time_stamp", pre=True)
    def validate_timestamp(cls, value: str) -> dt.datetime:
        """Parse the `time_stamp` attribute to a datetime

        Parameters
        ----------
        value : str
            Input value

        Returns
        -------
        dt.datetime
            Parsed datetime
        """

        #: pylint: disable=no-self-argument
        return dt.datetime.strptime(value, "%Y/%m/%d %H:%M:%S")

    @property
    def event_key(self) -> str:
        """The event key of a Direct Invocation event is its trigger

        Returns
        -------
        str
            Event key
        """
        return self.direct_invocation.trigger

    @classmethod
    def matches_event_type(cls, event: Dict[str, Any]) -> bool:
        """Infers whether the given event is a Direct Invocation event

        Parameters
        ----------
        event : Dict[str, Any]
            Lambda event

        Returns
        -------
        bool
            `True` if `event` is a Direct Invocation event
        """
        return "direct_invocation" in event



"""Defines the `Depends` class"""

import inspect

from typing import Callable, Generic, Iterator, Optional, TypeVar, Union
from typing_extensions import TypeGuard

__all__ = ["Dependency", "DependencyC"]

R = TypeVar("R")


def is_generator(
    func: Union[Callable[[], R], Callable[[], Iterator[R]]]
) -> TypeGuard[Callable[[], Iterator[R]]]:
    """Type guard to assert that `func` returns an `Iterator[R]`

    Parameters
    ----------
    func : Union[Callable[P, R], Callable[P, Iterator[R]]]
        Callable

    Returns
    -------
    TypeGuard[Callable[P, Iterator[R]]]
        `True` if `func` returns an `Iterator[R]`
    """
    return inspect.isgeneratorfunction(func)


def is_not_generator(
    func: Union[Callable[[], R], Callable[[], Iterator[R]]]
) -> TypeGuard[Callable[[], R]]:
    """Type guard to assert that `func` returns an `R`

    Parameters
    ----------
    func : Union[Callable[P, R], Callable[P, Iterator[R]]]
        Callable

    Returns
    -------
    TypeGuard[Callable[P, R]]
        `True` if `func` returns an `R`
    """
    return not inspect.isgeneratorfunction(func)


def get_result(
    func: Union[Callable[[], R], Callable[[], Iterator[R]]],
) -> R:
    """Get the result of calling a function; if necessary with `next`

    Parameters
    ----------
    func : Union[Callable[P, R], Callable[P, Iterator[R]]]
        Callable
    *args : Any
        Arguments of `func`
    **kwargs : Any
        Keyword arguments of `func`

    Returns
    -------
    R
        Return value of `func`

    Raises
    ------
    ValueError
        If we cannot get a result
    """
    if is_generator(func):
        return next(func())
    if is_not_generator(func):
        return func()
    raise ValueError("Cannot get here!")


class DependencyC(Generic[R]):
    """Class for defining dependencies

    Parameters
    ----------
    dependency : Union[Callable[P, R], Callable[P, Iterator[R]]]
        A function or generator dependency
    use_cache : bool, default `True`
        Whether to cache the result of calling `dependency`
    """

    def __init__(
        self,
        dependency: Union[Callable[[], R], Callable[[], Iterator[R]]],
        *,
        use_cache: bool = True,
    ) -> None:
        self.dependency = dependency
        self.use_cache = use_cache
        self.result: Optional[R] = None

    def __call__(self) -> R:
        if self.result:
            return self.result
        result = get_result(self.dependency)
        if self.use_cache:
            self.result = result
        return result


def Dependency(
    dependency: Union[Callable[[], R], Callable[[], Iterator[R]]],
    *,
    use_cache: bool = True,
) -> DependencyC[R]:
    """Create a Dependency instance"""
    return DependencyC(dependency=dependency, use_cache=use_cache)



"""Base AWS events models"""

from __future__ import annotations

import json

#: pylint: disable=no-name-in-module
from pydantic import ConstrainedStr, Field
from pydantic.generics import GenericModel

from typing import Any, Dict, Generic, Optional, TypeVar

#: pylint: enable=no-name-in-module


__all__ = [
    "ApiGatewayEvent",
    "AwsAccountNumber",
    "BranchBaseModel",
    "BranchGenericModel",
    "DataT",
    "UnparseableResposeDictionary",
    "LambdaResponse",
]


def to_camel(name: str) -> str:
    """Convert a string from snake case to camel case

    Parameters
    ----------
    name : str
        String to convert, probably in `snake_case`

    Returns
    -------
    str
        Result, in `camelCase`
    """
    first, *others = name.split("_")
    remainder = "".join(map(str.title, others))
    return first + remainder


class AwsAccountNumber(ConstrainedStr):
    """AWS Account Number"""

    min_length = 12
    max_length = 12


DataT = TypeVar("DataT")


class BranchGenericModel(GenericModel, Generic[DataT]):
    """Generic Pydantic model with standard configuration"""

    class Config:
        """Model configuration"""

        alias_generator = to_camel
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        extra = "allow"
        underscore_attrs_are_private = True


BranchBaseModel = BranchGenericModel[Any]


class AwsEvent(BranchGenericModel, Generic[DataT]):
    """Generic Pydantic model for *event* models"""

    @property
    def event_key(self) -> str:
        """The event key for this particular event type

        Returns
        -------
        str
            Event key

        Raises
        ------
        NotImplementedError
            In the base class
        """
        raise NotImplementedError()

    @classmethod
    def matches_event_type(cls, event: Dict[str, Any]) -> bool:
        """Infers whether the given event is of the type handled
        by this class

        Parameters
        ----------
        event : Dict[str, Any]
            Lambda event

        Returns
        -------
        bool
            `True` if this class handles the event `event`

        Raises
        ------
        NotImplementedError
            In the base class
        """
        raise NotImplementedError()


class ApiGatewayEvent(AwsEvent):
    """API Gateway API Proxy event; this model should not be used
    directy and isn't included in the docs; it's only used for
    identifying these event types any passing them to Mangum

    Parameters
    ----------
    event : Dict[str, Any]
        Unparsed event
    """

    __root__: Dict[str, Any] = Field(description="Unparsed event")

    @property
    def event_key(self) -> str:
        """Implement `event_key`, even though we will never call it

        Returns
        -------
        str
            Empty string
        """
        return ""

    @classmethod
    def matches_event_type(cls, event: Dict[str, Any]) -> bool:
        """Infers whether the given event is an API Gateway event

        Parameters
        ----------
        event : Dict[str, Any]
            Lambda event

        Returns
        -------
        bool
            `True` if `event` is an API Gateway event
        """
        return "pathParameters" in event


class UnparseableResposeDictionary(Exception):
    """Raised when a `LambdaResponse` cannot parse a response dictionary"""

    pass


class LambdaResponse:
    """A model for a Lambda function response

    Parameters
    ----------
    status_code
        Response status code
    body : Optional[Dict[str, Any]] = None
        Response body, by default `None`; should be a dictionary
    headers : Optional[Dict[str, Any]] = None
        Response headers, by default `None`
    """

    DEFAULT_HEADERS = {
        "Access-Control-Allow-Origin": "*",
        "Access-Control-Allow-Headers": "*",
        "Access-Control-Allow-Credentials": True,
        "Content-Type": "application/json",
        "X-Requested-With": "*",
    }

    def __init__(
        self,
        status_code: str,
        body: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ):
        self.is_base64_encoded = False
        self.status_code = status_code
        self.body = body
        self.headers = headers or self.DEFAULT_HEADERS

    @property
    def raw_body(self) -> str:
        """Raw response body

        Returns
        -------
        Response body as a JSON string
        """
        if self.body is not None:
            return json.dumps(self.body)
        return ""

    def dict(self) -> Dict[str, Any]:
        """Convert the instance to a dictionary for passing back the response

        Returns
        -------
        Dict[str, Any]
            Response dictionary
        """

        return {
            "isBase64Encoded": self.is_base64_encoded,
            "statusCode": self.status_code,
            "headers": self.headers,
            "body": self.raw_body,
        }

    @classmethod
    def parse_obj(cls, obj: Dict[str, Any]) -> "LambdaResponse":
        """Parse an object as a `LambdaResponse` instance

        Parameters
        ----------
        obj : Dict[str, Any]
            Object to parse

        Returns
        -------
        LambdaResponse
            `LambdaResponse` instance

        Raises
        ------
        ValueError
            If `obj` is not a dictionary
        UnparseableResposeDictionary
            If `obj` does not have a `statusCode` key
        """
        if not isinstance(obj, dict):
            raise ValueError("Cannot parse non-dictionary objects!")
        try:
            return cls(
                status_code=obj["statusCode"],
                body=obj.get("body"),
                headers=obj.get("headers"),
            )
        except KeyError:
            raise UnparseableResposeDictionary() from None



"""SNS Events data model"""

from __future__ import annotations

import datetime as dt

#: pylint: disable=no-name-in-module
from pydantic import BaseModel, Field
from pydantic.networks import HttpUrl

from .base import AwsEvent, BranchGenericModel, DataT

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
)

__all__ = [
    "SnsMessage",
    "SnsMsgAttributeModel",
    "SnsRecord",
    "SnsEvent",
]


if TYPE_CHECKING:
    J = TypeVar("J")

    class Json(Generic[J]):
        """Fake `Json` class to keep mypy happy"""

        def __class_getitem__(cls, item: Type[J]) -> Type[J]:
            #: pylint: disable=unused-argument
            ...

else:
    from pydantic import Json


class SnsMsgAttributeModel(BaseModel):
    """Message attributes

    Parameters
    ----------
    type : str
        Attribute type
    value : str
        Attribute value
    """

    type: str = Field(alias="Type")
    value: str = Field(alias="Value")


class SnsMessage(BranchGenericModel, Generic[DataT]):
    """An SNS message

    Parameters
    ----------
    timestamp : dt.datetime
        Timestamp
    message : str
        Message
    """

    subject: Optional[str]
    topic_arn: str
    unsubscribe_url: HttpUrl
    type: Literal["Notification"]
    message_attributes: Optional[Dict[str, SnsMsgAttributeModel]]
    message: Union[Json[DataT], str]
    message_id: str
    signing_cert_url: HttpUrl
    signature: str
    timestamp: dt.datetime
    signature_version: str

    class Config:
        """Model configuration"""

        alias_generator = lambda s: "".join(map(str.title, s.split("_")))  # noqa: E731


class SnsRecord(BranchGenericModel, Generic[DataT]):
    """An SNS record

    Parameters
    ----------
    event_source : Literal["aws:sns"]
        Event source; must be `"aws:sns"`
    event_subscription_arn : str
        Event subscription ARN
    sns : SnsMessage
        Message instance
    """

    event_source: Literal["aws:sns"] = Field(alias="EventSource")
    event_version: str = Field(alias="EventVersion")
    event_subscription_arn: str = Field(alias="EventSubscriptionARN")
    sns: SnsMessage[DataT] = Field(alias="Sns")


class SnsEvent(AwsEvent, Generic[DataT]):
    """An SNS Event

    Parameters
    ----------
    records : List[SnsRecord]
        List of `SqsRecord` instances
    """

    records: List[SnsRecord[DataT]] = Field(alias="Records")

    @property
    def topic_name(self) -> str:
        """Topic name, defined as the event subscription ARN of the first record

        Returns
        -------
        str
            Topic name
        """
        return self.records[0].event_subscription_arn.split(":")[-1]

    @property
    def event_key(self) -> str:
        """Event key, defined as the topic name

        Returns
        -------
        str
            Topic name
        """
        return self.topic_name

    @classmethod
    def matches_event_type(cls, event: Dict[str, Any]) -> bool:
        """Infers whether the given event is an SNS event

        Parameters
        ----------
        event : Dict[str, Any]
            Lambda event

        Returns
        -------
        bool
            `True` if `event` is an SNS event
        """

        if "Records" in event:
            records = event["Records"]
            if isinstance(records, list):
                return records[0].get("Sns") is not None
        return False



"""SQS Events data model"""

from __future__ import annotations

import datetime as dt

#: pylint: disable=no-name-in-module
from pydantic import Field

from .base import AwsEvent, BranchBaseModel, BranchGenericModel, DataT

from typing import Any, Dict, Generic, List, Literal, Optional

#: pylint: enable=no-name-in-module


__all__ = ["SqsRecordAttributes", "SqsMesssageAttributes", "SqsRecord", "SqsEvent"]


## SQS


class SqsRecordAttributes(BranchBaseModel):
    """SQS Record attributes

    Parameters
    ----------
    approximate_recieve_count : str
        Approximate received count
    approximate_first_receive_timestamp : dt.datetime
        Approximate first received timestamp
    message_deduplication_id : Optional[str]
        Message deduplication ID
    message_group_id : Optional[str]
        Message group ID
    sender_id : str
        Sender ID
    sent_timestamp : dt.datetime
        Sent timestamp
    sequence_number : Optional[str]
        Sequence number
    aws_trace_header : Optional[str]
        AWS trace header
    """

    approximate_receive_count: str
    approximate_first_receive_timestamp: dt.datetime
    message_deduplication_id: Optional[str]
    message_group_id: Optional[str]
    sender_id: str
    sent_timestamp: dt.datetime
    sequence_number: Optional[str]
    aws_trace_header: Optional[str] = Field(alias="AWSTraceHeader")

    class Config:
        """Model configuration"""

        alias_generator = lambda s: "".join(map(str.title, s.split("_")))  # noqa: E731


class SqsMesssageAttributes(BranchBaseModel):
    """SQS Message attributes

    Parameters
    ----------
    string_value : Optional[str]
        String value
    binary_value : Optional[str]
        Binary value
    string_list_values : List[str] = []
        List of string values
    binary_list_values : List[str] = []
        List of binary values
    data_type : str
        Data type
    """

    string_value: Optional[str]
    binary_value: Optional[str]
    string_list_values: List[str] = []
    binary_list_values: List[str] = []
    data_type: str


class SqsRecord(BranchGenericModel, Generic[DataT]):
    """An SQS Record

    Parameters
    ----------
    message_id : str
        Message ID
    receipt_handle : str
        Receipt handle
    body : str
        Message body
    attributes : SqsRecordAttributes
        Record attributes
    message_attributes : Dict[str, SqsMessageAttributes]
        Message attributes
    md5_of_body : str
        Body MD5 hash
    md5_of_message_attributes : Optional[str]
        Message attributes MD5 hash
    event_source : Literal["aws:sqs"]
        Event source; must be `"aws:sqs"`
    event_source_arn : str
        Event source ARN: the resource that generated the event
    aws_region : str
        AWS region
    """

    message_id: str
    receipt_handle: str
    body: DataT
    attributes: SqsRecordAttributes
    message_attributes: Dict[str, SqsMesssageAttributes]
    md5_of_body: str
    md5_of_message_attributes: Optional[str]
    event_source: Literal["aws:sqs"]
    event_source_arn: str = Field(alias="eventSourceARN")
    aws_region: str


class SqsEvent(AwsEvent, Generic[DataT]):
    """An SQS Event

    Parameters
    ----------
    records : List[SqsRecord]
        List of `SqsRecord` instances
    """

    records: List[SqsRecord[DataT]] = Field(alias="Records")

    @property
    def queue_name(self) -> str:
        """Queue name, defined as the ARN of the first record's event source

        Returns
        -------
        str
            Queue name
        """
        return self.records[0].event_source_arn.split(":")[-1]

    @property
    def event_key(self) -> str:
        """Event key, defined as the queue name

        Returns
        -------
        str
            Queue name
        """
        return self.queue_name

    @classmethod
    def matches_event_type(cls, event: Dict[str, Any]) -> bool:
        """Infers whether the given event is an SQS event

        Parameters
        ----------
        event : Dict[str, Any]
            Lambda event

        Returns
        -------
        bool
            `True` if `event` is an SQS event
        """

        if "Records" in event:
            records = event["Records"]
            if isinstance(records, list):
                return records[0].get("eventSource") == "aws:sqs"
        return False



"""Utility functions"""

import inspect
import logging
from functools import wraps

from lambda_handler.dependencies import DependencyC

#: pylint: disable=duplicate-code
from lambda_handler.model import (
    ApiGatewayEvent,
    AwsEvent,
    DirectInvocationEvent,
    EventBridgeEvent,
    LambdaResponse,
    SnsEvent,
    SqsEvent,
)
from lambda_handler.types import (
    AwsEventCallable,
    AwsEventRawCallable,
    AwsEventType,
    LambdaHandlerInterface,
)

from typing import Any, Callable, Dict, List, Literal, Type, TypeVar, Union, overload
from typing_extensions import Concatenate, ParamSpec, TypeAlias

#: pylint: enable=duplicate-code


__all__ = [
    "AwsEventCallable",
    "OnetimeDictionary",
    "EventKeyMismatch",
    "validate_function_signature",
    "parse_lambda_event",
]


Bool: TypeAlias = Union[Literal[True], Literal[False]]


_KT = TypeVar("_KT")
_VT = TypeVar("_VT")


class OnetimeDictionary(dict):
    """A dictionary which only allows a key to be set once"""

    def __setitem__(self, __k: _KT, __v: _VT) -> None:
        if __k in self:
            raise ValueError(f"The key {__k} has already been registered!")
        return super().__setitem__(__k, __v)


class EventKeyMismatch(Exception):
    """Raised when a function is passed an event key for which it is not registered"""

    pass


class UnknownEventType(Exception):
    """Raised when an event is of an unknown type"""

    pass


def validate_function_signature(func: Callable, input_type: Type[AwsEvent]) -> bool:
    """Valiate the given `func` has the correct signature

    Parameters
    ----------
    func : Callable
        Function that _should_ accept a single argument of type `input_type`
        and return a `LambdaResponse`
    input_type : Type[AwsEvent]
        Input type to validate the first argument of `func`

    Returns
    -------
    bool
        `True`; otherwise raises an exception

    Raises
    ------
    ValueError
        If `func` does not have the correct signature
    """

    sig = inspect.signature(func)
    params = [
        param
        for param in sig.parameters.values()
        if not isinstance(param.default, DependencyC)
    ]

    func_name = func.__qualname__

    if not sig.return_annotation == LambdaResponse:
        raise ValueError(f"`{func_name}` does not return a `LambdaResponse` instance!")
    if not len(params) == 1:
        raise ValueError(
            f"`{func_name}` should only accept a single value besides its dependencies!"
        )
    if not issubclass(params[0].annotation, input_type):
        print(params[0].annotation)
        raise ValueError(
            f"`{func_name}` should accept a single parameter of type `{input_type}`"
        )
    return True


_P = ParamSpec("_P")


def create_raw_outer(
    handler: LambdaHandlerInterface,
    event_type: Type[AwsEventType],
    event_key: str,
) -> Callable[
    [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
    AwsEventRawCallable,
]:
    """Create a wrapper for a function that accepts a `Dict[str, Any]`
    plus dependencies, and returns a `Dict[str, Any]`

    Parameters
    ----------
    handler : LambdaHandlerInterface
        LambdaHandlerInterface instance
    event_type : Type[AwsEvent[T]]
        Event type to create the wrapper for
    event_key : str
        Event key

    Returns
    -------
    Callable[
        [Callable[Concatenate[Dict[str, Any], P], Dict[str, Any]]],
        Callable[[Dict[str, Any]], Dict[str, Any]],
    ]
        A wrapper for a function that accepts a `Dict[str, Any]`
    plus dependencies, and returns a `Dict[str, Any]`
    """

    def outer(
        func: Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]
    ) -> AwsEventRawCallable:
        @wraps(func)
        def inner(event: Dict[str, Any]) -> Dict[str, Any]:
            sig = inspect.signature(func)
            dependencies: Dict[str, Any] = {
                k: p.default()
                for k, p in sig.parameters.items()
                if isinstance(p.default, DependencyC)
            }
            return func(event, **dependencies)

        handler.add_event_func(event_type, event_key, inner)

        return inner

    return outer


def create_typed_outer(
    handler: LambdaHandlerInterface,
    event_type: Type[AwsEventType],
    event_key: str,
) -> Callable[
    [Callable[Concatenate[AwsEventType, _P], LambdaResponse]],
    AwsEventRawCallable,
]:
    """Create the wrapper function for a typed callable

    Parameters
    ----------
    handler : LambdaHandlerInterface
        `LambdaHandlerInterface` instance
    event_type : Type[AwsEventType]
        Event type
    event_key : str
        Attribute name of the event type used for routing

    Returns
    -------
    Callable[ [Callable[Concatenate[AwsEventType, _P], LambdaResponse]], AwsEventRawCallable, ]
        _description_
    """

    def outer(
        func: Callable[Concatenate[AwsEventType, _P], LambdaResponse]
    ) -> Callable[[Dict[str, Any]], Dict[str, Any]]:
        @wraps(func)
        def inner(event: Dict[str, Any]) -> Dict[str, Any]:
            """Parses the event to the correct type and passes it to `func`

            Parameters
            ----------
            event : Dict[str, Any]
                Unprocessed event

            Returns
            -------
            LambdaResponse
                `LambdaResponse` instance
            """
            # Get the actual event type from the annotation
            sig = inspect.signature(func)
            params = list(sig.parameters.values())
            concrete_event_type = params[0].annotation

            dependencies: Dict[str, Any] = {
                k: p.default()
                for k, p in sig.parameters.items()
                if isinstance(p.default, DependencyC)
            }

            parsed_event = concrete_event_type.parse_obj(event)
            if parsed_event.event_key != event_key:
                raise EventKeyMismatch(
                    f"Event key mismatch! `'{parsed_event.event_key}' != '{event_key}'`"
                )
            return func(parsed_event, **dependencies).dict()

        handler.add_event_func(event_type, event_key, inner)

        return inner

    return outer


@overload
def create_wrapper(
    handler: LambdaHandlerInterface,
    event_type: Type[AwsEventType],
    event_key: str,
    raw: Literal[True],
) -> Callable[
    [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
    AwsEventRawCallable,
]:
    pass


@overload
def create_wrapper(
    handler: LambdaHandlerInterface,
    event_type: Type[AwsEventType],
    event_key: str,
    raw: Literal[False],
) -> Callable[
    [Callable[Concatenate[AwsEventType, _P], LambdaResponse]],
    AwsEventRawCallable,
]:
    pass


@overload
def create_wrapper(
    handler: LambdaHandlerInterface,
    event_type: Type[AwsEventType],
    event_key: str,
    raw: Bool,
) -> Union[
    Callable[
        [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
        AwsEventRawCallable,
    ],
    Callable[
        [Callable[Concatenate[AwsEventType, _P], LambdaResponse]],
        AwsEventRawCallable,
    ],
]:
    pass


def create_wrapper(
    handler: LambdaHandlerInterface,
    event_type: Type[AwsEventType],
    event_key: str,
    raw: Bool = True,
) -> Union[
    Callable[
        [Callable[Concatenate[Dict[str, Any], _P], Dict[str, Any]]],
        AwsEventRawCallable,
    ],
    Callable[
        [Callable[Concatenate[AwsEventType, _P], LambdaResponse]],
        AwsEventRawCallable,
    ],
]:
    """Create a wrapper for a given event type

    Parameters
    ----------
    handler : LambdaHandlerInterface
        Handler instance
    event_type : Type[AwsEvent]
        Event type to create the wrapper for
    event_key : str
        Event key (topic, queue, etc.) name for storing the callable
    raw : bool = `False`
        Don't parse the event before handing it to the function, and
        assume that the function returns

    Returns
    -------
    Union[
        Callable[[AwsEventTypedCallable], AwsEventCallable],
        Callable[[AwsEventRawCallable], AwsEventRawCallable]
    ]
        A callable that wraps a callable of a given signature and returns
        it
    """

    if raw:
        return create_raw_outer(
            handler=handler,
            event_type=event_type,
            event_key=event_key,
        )
    return create_typed_outer(
        handler=handler,
        event_type=event_type,
        event_key=event_key,
    )


def parse_lambda_event(event: Dict[str, Any]) -> AwsEvent:
    """Parses a Dictionary from an lambda event to an `AwsEvent` type

    Parameters
    ----------
    event : Dict[str, Any]
        Event to parse

    Returns
    -------
    AwsEvent
        Parsed event

    Raises
    ------
    UnknownEventType
        If it cannot parse `event`
    """
    event_types: List[Type[AwsEvent]] = [
        ApiGatewayEvent,
        DirectInvocationEvent,
        EventBridgeEvent,
        SnsEvent,
        SqsEvent,
    ]

    for event_type in event_types:
        if event_type.matches_event_type(event):
            name = event_type.__qualname__
            logging.info("Found %s event type", name)
            return event_type.parse_obj(event)

    raise UnknownEventType(f"Cannot parse event `{event}`!")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lambda_handler', 'lambda_handler.model']

package_data = \
{'': ['*']}

modules = \
['py']
install_requires = \
['pydantic>=1.10.2,<2.0.0']

setup_kwargs = {
    'name': 'lambda-handler',
    'version': '0.1.0',
    'description': 'A Python package for routing and validating AWS events inside a Lambda function',
    'long_description': None,
    'author': 'Matthew Badger',
    'author_email': 'matt@branchenergy.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""AWS event models"""

from .base import (  # noqa: F401
    ApiGatewayEvent,
    AwsEvent,
    DataT,
    LambdaResponse,
    UnparseableResposeDictionary,
)
from .direct_invocation import DirectInvocationDetail  # noqa: F401
from .direct_invocation import DirectInvocationEvent
from .event_bridge import EventBridgeEvent, EventBridgeResource  # noqa: F401
from .sns import SnsEvent, SnsMessage, SnsMsgAttributeModel, SnsRecord  # noqa: F401
from .sqs import (  # noqa: F401
    SqsEvent,
    SqsMesssageAttributes,
    SqsRecord,
    SqsRecordAttributes,
)
